#include "kanji.h"

/*
 * kgetc():	get character sequence from file stream and
 *          returns them in 16-bit format.
 */

#define KANJI_STATE	0xff00
#define NORMAL		0x0000
#define KANJI		0x0100
#define KANA		0x0200

#define ESC_STATE	0x00ff
#define ESC		0x0001
#define ESC_DOL		0x0002
#define ESC_PAR		0x0003

#define putq(c)		(kp->ks_queue[kp->ks_ql++] = (c))

int
kgetc(kp)
  register KSTREAM *kp;
{
  register int cc;
  int c1;
  register int n;
  char buf[BUFSIZ];
  register char *p;

  for (;;) {
    if (kp->ks_ql > 0) {
      cc = kp->ks_queue[kp->ks_qp++];
      if (kp->ks_qp >= kp->ks_ql) kp->ks_ql = 0;
      return cc;
    }
    kp->ks_qp = 0;

    if (!(n = (*kp->ks_getf)(kp->ks_id, buf, BUFSIZ))) {
      /* nothing to read. reset escaping status */
      if (KS_INTERP(kp->ks_flag) == KS_KANJI &&
	  KS_CODE(kp->ks_flag) == KS_JIS) {
	switch (kp->ks_gstate & ESC_STATE) {
	case ESC:
	  putq(ESCAPE);
	  break;
	case ESC_PAR:
	  putq(ESCAPE);
	  putq('(');
	  break;
	case ESC_DOL:
	  putq(ESCAPE);
	  putq('$');
	  break;
	}
	kp->ks_gstate &= ~ESC_STATE;
      }
      continue;
    }
    if (n < 0) return EOF;
    for (p = buf; n-- > 0; ) {
      if (KS_INTERP(kp->ks_flag) == KS_BINARY) {
	putq(*p++ & (KS_THRU(kp->ks_flag)?0xff: 0x7f));
	continue;
      }
      cc = *p++;
      if (!KS_THRU(kp->ks_flag) ||
	  (KS_INTERP(kp->ks_flag) == KS_KANJI &&
	   KS_CODE(kp->ks_flag) == KS_JIS)) {
	/* input characters are stripped to 7 bits */
	cc &= 0177;
      } else
	cc &= 0xff;
      if (c1 = kp->ks_buf[0]) {
	kp->ks_buf[0] = 0;
	/* combine broken 2-byte character */
	switch (KS_CODE(kp->ks_flag)) {
	case KS_JIS:
	  if (cc > ' ' && cc < 0x7f) {
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    enkanji(cc, c1, cc);
	    putq(cc);
	    continue;
	  }
	  break;
	case KS_UJIS:
	  if (c1 == SSHFT2 && cc & 0x80 &&
	      (cc & 0x7f) > ' ' && (cc & 0x7f)< 0x7f) {
	    /* UJIS KANA */
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    enkana(cc, cc);
	    putq(cc);
	    continue;
	  }
	  if ((cc & 0x7f) > ' ' && (cc & 0x7f) < 0x7f) {
	    /* UJIS KANJI */
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    enkanji(cc, c1&0x7f, cc&0x7f);
	    putq(cc);
	    continue;
	  }
	  break;
	case KS_SJIS:
	  if (cc >= 0x40) {
	    int c2 = (c1 << 8) | cc;

	    if (stoj(&c2, &c2)) break;
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    enkanji(cc, (c2>>8)&0x7f, c2&0x7f);
	    putq(cc);
	    continue;
	  }
	  break;
	}
	putq(c1);
      }
      switch (kp->ks_gstate) {
      case NORMAL:
      case KANJI:
      case KANA:
      case KANJI | KANA:
	switch (cc) {
	case ESCAPE:
	  if (KS_CODE(kp->ks_flag) == KS_JIS ||
	      KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	    KS_CODE(kp->ks_flag) = KS_JIS;
	    kp->ks_gstate += ESC;
	    continue;
	  }
	  /* no special meaning */
	  break;
	case SHFTOU:
	  if (KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	    /* treat as JIS KANJI */
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    KS_CODE(kp->ks_flag) = KS_JIS;
	    KS_PASS(kp->ks_flag) = 0;
	  }
	  if (KS_CODE(kp->ks_flag) == KS_JIS && !KS_PASS(kp->ks_flag)) {
	    kp->ks_gstate |= KANA;
	    continue;
	  }
	  /* no special meaning */
	  break;
	case SHFTIN:
	  if (KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	    /* treat as JIS KANJI */
	    KS_INTERP(kp->ks_flag) = KS_KANJI;
	    KS_CODE(kp->ks_flag) = KS_JIS;
	    KS_PASS(kp->ks_flag) = 0;
	  }
	  if (KS_CODE(kp->ks_flag) == KS_JIS && !KS_PASS(kp->ks_flag)) {
	    kp->ks_gstate &= ~KANA;
	    continue;
	  }
	  /* no special meaning */
	  break;
	case SSHFT2:
	  if (KS_CODE(kp->ks_flag) == KS_UJIS ||
	      KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	    KS_CODE(kp->ks_flag) = KS_UJIS;
	    if (n < 1) {
	      /* broken */
	      kp->ks_buf[0] = cc;
	      continue;
	    }
	    c1 = *p++;
	    if (c1 & 0x80 && !iscntrl(c1 & 0x7f)) {
	      KS_INTERP(kp->ks_flag) = KS_KANJI;
	      n--;
	      enkana(cc, c1 & 0x7f);
	      putq(cc);
	      continue;
	    }
	    p--;
	  }
	  /* no special meaning */
	  break;
	}
	if (cc & 0x80) {
	  if (cc < 0xa0) {
	    /* may be 1st byte of SJIS KANJI */
	    if (KS_CODE(kp->ks_flag) == KS_SJIS ||
		KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	      KS_CODE(kp->ks_flag) = KS_SJIS;
	      if (n < 1) {
		/* broken */
		kp->ks_buf[0] = cc;
		continue;
	      }
	      c1 = *p++ & 0xff;
	      if (!iscntrl(c1) && ((c1 |= (cc << 8)), !stoj(&c1, &c1))) {
		n--;
		KS_INTERP(kp->ks_flag) = KS_KANJI;
		enkanji(cc, (c1>>8)&0x7f, c1&0x7f);
		putq(cc);
		continue;
	      }
	      p--;
	    }
	  } else if (cc < 0xe0) {
	    /*
	     * may be 1st byte of UJIS KANJI
	     * or KANA of SJIS
	     */
	    if (KS_CODE(kp->ks_flag) == KS_UJIS ||
		KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	      KS_CODE(kp->ks_flag) = KS_UJIS;
	      if (n < 1) {
		/* broken */
		kp->ks_buf[0] = cc;
		continue;
	      }
	      c1 = *p++ & 0xff;
	      if (c1 & 0x80 && (c1 & 0x7f) > ' ' && (c1 & 0x7f) < 0x7f) {
		KS_INTERP(kp->ks_flag) = KS_KANJI;
		n--;
		enkanji(cc, cc&0x7f, c1&0x7f);
		putq(cc);
		continue;
	      }
	      p--;
	    }
	    if (KS_CODE(kp->ks_flag) == KS_SJIS ||
		KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	      KS_CODE(kp->ks_flag) = KS_SJIS;
	      KS_INTERP(kp->ks_flag) = KS_KANJI;
	      enkana(cc, cc&0x7f);
	      putq(cc);
	      continue;
	    }
	  } else {
	    /*
	     * may be UJIS kanji or SJIS kanji
	     */
	    if (KS_CODE(kp->ks_flag) == KS_UJIS ||
		KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	      KS_CODE(kp->ks_flag) = KS_UJIS;
	      if (n < 1) {
		/* broken */
		kp->ks_buf[0] = cc;
		continue;
	      }
	      c1 = *p++ & 0xff;
	      if (c1 & 0x80 && (c1 & 0x7f) > ' ' && (c1 & 0x7f) < 0x7f) {
		KS_INTERP(kp->ks_flag) = KS_KANJI;
		n--;
		enkanji(cc, cc&0x7f, c1&0x7f);
		putq(cc);
		continue;
	      }
	      p--;
	    }
	    if (KS_CODE(kp->ks_flag) == KS_SJIS ||
		KS_INTERP(kp->ks_flag) == KS_UKANJI) {
	      KS_CODE(kp->ks_flag) = KS_SJIS;
	      if (n < 1) {
		/* broken */
		kp->ks_buf[0] = cc;
		continue;
	      }
	      c1 = *p++ & 0xff;
	      if (!iscntrl(c1) && ((c1 |= (cc << 8)), !stoj(&c1, &c1))) {
		n--;
		KS_INTERP(kp->ks_flag) = KS_KANJI;
		enkanji(cc, (c1>>8)&0x7f, c1&0x7f);
		putq(cc);
		continue;
	      }
	      p--;
	    }
	  }
	}
	if (iscntrl(cc)) {
	  putq(cc);
	  continue;
	}
	if (KS_CODE(kp->ks_flag) == KS_JIS && cc > ' ' && cc < 0x7f) {
	  if (kp->ks_gstate & KANA) {
	    /* Kana precedes Kanji */
	    enkana(cc, cc&0x7f);
	    putq(cc);
	    continue;
	  }
	  if (kp->ks_gstate & KANJI) {
	    /* get second byte */
	    if (n < 1) {
	      /* kanji code is broken. */
	      kp->ks_buf[0] = cc;
	      continue;
	    }
	    c1 = *p++ & 0x7f; /* !THRU assumed */
	    if (c1 > ' ' && c1 < 0x7f) {
	      n--;
	      enkanji(cc, cc, c1);
	      putq(cc);
	      continue;
	    }
	    p--;
	  }
	}
	/* Alphabet or uncategolized (META) characters */
	putq(cc);
	continue;
      case ESC:
      case ESC | KANJI:
      case ESC | KANA:
      case ESC | KANJI | KANA:
	/* got ESC */
	kp->ks_gstate -= ESC;
	if (cc == '$') {
	  kp->ks_gstate += ESC_DOL;
	  continue;
	}
	if (cc == '(') {
	  kp->ks_gstate += ESC_PAR;
	  continue;
	}
	/* other ESC sequence. pass through */
	putq(ESCAPE);
	putq(cc);
	continue;
	
      case ESC_DOL:
      case ESC_DOL | KANJI:
      case ESC_DOL | KANA:
      case ESC_DOL | KANJI | KANA:
	/* got ESC $ */
	kp->ks_gstate -= ESC_DOL;
	if (cc == '@' || cc == 'B') {
	  KS_INTERP(kp->ks_flag) = KS_KANJI;
	  kp->ks_gstate |= KANJI;
	  KS_KI(kp->ks_flag) = (cc == '@')? KS_OLDJIS: KS_NEWJIS;
	  continue;
	}
	/* other ESC sequence. pass through */
	putq(ESCAPE);
	putq('$');
	putq(cc);
	continue;
      case ESC_PAR:
      case ESC_PAR | KANJI:
      case ESC_PAR | KANA:
      case ESC_PAR | KANJI | KANA:
	/* got ESC ( */
	kp->ks_gstate -= ESC_PAR;
	if (cc == 'A' || cc == 'B' || cc == 'J' || cc == 'H') {
	  KS_INTERP(kp->ks_flag) = KS_KANJI;
	  kp->ks_gstate &= ~KANJI;
	  KS_RI(kp->ks_flag) = (cc <= 'B')? KS_ASCII: KS_ROMAJI;
	  continue;
	}
	/* other ESC sequence. pass through */
	putq(ESCAPE);
	putq('(');
	putq(cc);
	continue;
      }
    }
  }
  /*NOTREACHED*/
}
