#include "kanji.h"
#include "../econfig.h"
#if BSD
# include <sys/ioctl.h>
# include <sys/time.h>
# include <sys/errno.h>
#endif
#if USG
# include <signal.h>
# include <errno.h>
#endif

/*
 * open Kanji stream on opened file stream fp.
 * kpass specifies if passes shift-in/out through.
 * programs which needs to catch ^N/^O must set it.
 */

struct KF {
	FILE *fp;			/* associated FILE pointer */
#if BSD
	struct timeval tm;	/* ESC timeout (passed to select) */
#endif
#if USG
	unsigned tm;		/* passed to alarm */
	int (*ac)();		/* last alarm handler */
	unsigned lt;		/* last alarm count rest */
#endif
	int lastesc;		/* last input was ESC */
};

KFILE *
kopen(fp, flag, totime)
	FILE *fp;
	unsigned flag;
	unsigned totime;
{
	register KSTREAM *kp;
	register struct KF *kfp;
	extern char *malloc();
	int f_open(), f_close(), f_get(), f_put();

	if (!fp
	    || !(kfp = (struct KF *)malloc(sizeof(struct KF)))) return NULL;
	kfp->fp = fp;
#if BSD
	kfp->tm.tv_sec = totime/1000;
	kfp->tm.tv_usec = (totime % 1000) * 1000;
#endif
#if USG
	kfp->tm = totime? 1: 0;
#endif
	if (!(kp = kalloc((caddr_t)kfp, f_open, f_close, f_get, f_put, flag)))
		return NULL;
	return kp;
}

/*ARGSUSED*/
static int
f_open(id)
	struct KF *id;
{
	id->lastesc = 0;
	return 0;
}

static int
f_close(id)
	struct KF *id;
{
	return fflush(id->fp);
}

static int
f_get(id, buf, len)
	register struct KF *id;
	char buf[];
	int len;
{
	register int n;
	extern int errno;

	if (id->lastesc) {
		/* check timeout */
#if BSD
		if (id->tm.tv_usec || id->tm.tv_sec) {
			int r;
		    retry:
			r = (1<<(fileno(id->fp)));
			n = select(32, &r, (int *)NULL, (int *)NULL, &id->tm);
			if (!n) return 0;
			if (n < 0) {
				if (errno == EINTR) goto retry;
				return -1;
			}
		}
#endif
#if USG
		if (id->tm) {
			int alarm_catcher();

			id->ac = signal(SIGALRM, alarm_catcher);
			id->lt = alarm(id->tm);
		}
#endif /*USG*/
	}
	n = read(fileno(id->fp), buf, len);
#if USG
	if (n < 0 && errno == EINTR) {
		/* timed out */
		(void)signal(SIGALRM, id->ac);
		(void)alarm(id->lt);
		return 0;
	}
#endif /*USG*/
	id->lastesc = (n > 0 && buf[n-1] == ESCAPE);
	return (n<=0)? -1: n;
}

#if USG
static int
alarm_catcher()
{
}
#endif

static int
f_put(id, buf, len)
	struct KF *id;
	char buf[];
	int len;
{
	(void)fwrite(buf, sizeof(char), len, id->fp);
	return len;
}
