#include "econfig.h"

/*** configuration parameters ***/

/* name, location, owner, group, permission of binary */
PROGRAM	      = kemacs
DEST	      = /usr/local/bin
OWNER         = root
GROUP         = staff
MODE	      = 755
/* name and location of manual */
MNAME	      = kemacs.1
MDEST	      = /usr/local/man/man1
/* location of library */
LIBDEST	      = /usr/local/lib/kemacs

/* Flags to CC, LD and INSTALL */
OPTIM	      =# -O
DEBUG	      = -g
DEFS	      =
#if BSD
#if HAS_AS_R
SHARESTR      = -R
#endif
STRIP	      = -s
#endif

/* File code for sources */
FCODE         = -j	/* JIS KANJI */
/* FCODE      = -u	/* UJIS KANJI */
/* FCODE      = -s	/* SJIS KANJI */

/* location of include file */
#if KANJI
INCL	      = -Ikanji -ICstrings
#else
INCL	      =
#endif

/* runtime libraries */
/*ELIBS	      = -ltermcap -lPW		# termcap + System V */
/*ELIBS	      = -lcurses -lPW		# System V */
#if BSD
ELIBS	      = -ltermlib		/* BSD */
#endif
#if USG_5_4
ELIBS         = -lcurses -lgen -lmalloc
#endif

/*** end of configuration part ***/

#if BSD
MAKE	      = exec "make" $(MFLAGS)
#endif
MAKE_HERE     = $(MAKE) -f make.file
MAKE_KANJI    = $(MAKE) -f Makefile.kanji CC='$(CC)' CCFLAGS='$(CCFLAGS)'
MAKE_KPP      = $(MAKE) -f Makefile.kpp CC='$(CC)' CCFLAGS='$(CCFLAGS)'
MAKE_CSTR     = $(MAKE) -f Makefile.Cstr CC='$(CC)' CCFLAGS='$(CCFLAGS)'
KPP	      = kpp/kpp
FTOC	      = kpp/ftoc
#if BSD
INSTALL	      = install
#endif
MAKETD	      = maketd
LINT	      = lint
CTAGS	      = ctags
#if USG
STRIP	      = strip
#endif

.SUFFIXES:
.SUFFIXES: .o .c .C .h .H

CCFLAGS        = $(OPTIM) $(DEBUG)
CFLAGS        = $(CCFLAGS) $(INCL) $(DEFS)
LDFLAGS	      = $(DEBUG) $(DEFS)
LINTFLAGS     = -xhb $(INCL)
#if	BSD
INSTFLAGS     = -c $(STRIP) -o $(OWNER) -g $(GROUP) -m $(MODE)
#endif

#if KANJI
LIBS          = kanji/kanji.a Cstrings/Cstrings.a
#else
LIBS	      =
#endif

OBJS	      = basic.o \
		bind.o \
		buffer.o \
		crypt.o \
		display.o \
		eval.o \
		exec.o \
		file.o \
		fileio.o \
		input.o \
		isearch.o \
		kanji.o \
		line.o \
		main.o \
		random.o \
		region.o \
		search.o \
		spawn.o \
		tcap.o \
		termio.o \
		version.o \
		window.o \
		word.o

SRCS	      = basic.C \
		bind.C \
		buffer.C \
		crypt.C \
		display.C \
		eval.C \
		exec.C \
		file.C \
		fileio.C \
		input.C \
		isearch.C \
		kanji.C \
		line.C \
		main.C \
		random.C \
		region.C \
		search.C \
		spawn.C \
		tcap.C \
		termio.C \
		version.C \
		window.C \
		word.C

CSRCS	      = basic.c \
		bind.c \
		buffer.c \
		crypt.c \
		display.c \
		eval.c \
		exec.c \
		file.c \
		fileio.c \
		input.c \
		isearch.c \
		kanji.c \
		line.c \
		main.c \
		random.c \
		region.c \
		search.c \
		spawn.c \
		tcap.c \
		termio.c \
		version.c \
		window.c \
		word.c

HDRS	      = ebind.H \
		ecomm.H \
		edef.H \
		efunc.H \
		epath.H \
		estruct.H \
		evar.H \
		eversion.H

HHDRS	      = ebind.h \
		ecomm.h \
		edef.h \
		efunc.h \
		epath.h \
		estruct.h \
		evar.h \
		eversion.h

MANUAL	      = kemacs.man

DOCS	      = Cmds.narrow Commands.me Macros Readme Tech WhatsNew \
		emacs.key emacs.hlp emacs.HLP emacs.tut emacs.TUT \
		emacs.mss emacs.MSS magic.doc Added Install Manifest

#if KANJI
STROBJ	      = Css.o
#else
STROBJ	      =
#endif

#if KANJI
all:		$(KPP) $(FTOC) $(PROGRAM)
#else
all:		$(PROGRAM)
#endif

debug:;       @ $(MAKE_HERE) all PROGRAM="a.out" \
			OPTIM="" \
			DEBUG="-g"

$(PROGRAM):     $(OBJS) $(LIBS) $(STROBJ)
#if BSD
	      @ echo -n "Loading $(PROGRAM) ..."
#endif
#if USG
	      @ echo "Loading $(PROGRAM) ...\c"
#endif
	      @ $(CC) $(LDFLAGS) $(STROBJ) $(OBJS) $(LIBS) $(ELIBS) -o $(PROGRAM)
	      @ echo "done"

clean:
	      @ echo "Cleaning up directory ..."
	      @ cd kpp; $(MAKE_KPP) clean
	      @ cd kanji; $(MAKE_KANJI) clean
	      @ cd Cstrings; $(MAKE_CSTR) clean
	      @ rm -f $(OBJS) $(CSRCS) $(HHDRS) Cxx.* Css.* a.out Cstr.data make.file *.bak core

realclean:
	      @ echo "Cleaning all binaries and executables ..."
	      @ cd kpp; $(MAKE_KPP) realclean
	      @ cd kanji; $(MAKE_KANJI) realclean
	      @ cd Cstrings; $(MAKE_CSTR) realclean
	      @ rm -f $(PROGRAM) $(OBJS) $(CSRCS) $(HHDRS) Cxx.* Css.* a.out Cstr.data make.file *.bak core

install:	inst-bin inst-man inst-lib

inst-bin:	$(PROGRAM)
#if BSD
	      @ echo -n Installing $(PROGRAM) in $(DEST) "..."
#endif
#if USG
	      @ echo Installing $(PROGRAM) in $(DEST) "...\c"
#endif
#if BSD
	      @ $(INSTALL) $(INSTFLAGS) $(PROGRAM) $(DEST)
#endif
#if USG
	      @ cp $(PROGRAM) $(DEST) && \
		cd $(DEST) && \
		$(STRIP) $(PROGRAM) && \
		chmod $(MODE) $(PROGRAM) && \
		chown $(OWNER) $(PROGRAM) && \
		chgrp $(GROUP) $(PROGRAM)
#endif
	      @ echo " done"

inst-man:
	      @ echo -n "Installing manual ..."
	      @ if [ `pwd`/$(MANUAL) != $(MDEST)/$(MNAME) ]; then \
			cp $(MANUAL) $(MDEST)/$(MNAME); \
			fi
	      @ echo " done"

inst-lib:
	      @ echo -n "Installing library ..."
	      @ if [ `pwd` != $(LIBDEST) ]; then \
			cp $(DOCS) $(LIBDEST); \
		fi
	      @ echo " done"

update:		$(DEST)/$(PROGRAM)

$(DEST)/$(PROGRAM): $(PROGRAM)
	      @ $(MAKE_HERE) install

lint:		$(CSRCS) $(HHDRS)
	      @ "$(LINT)" $(LINTFLAGS) $(CSRCS) 2>&1 | egrep -v 'possible pointer alignment'

tags:           $(HHDRS) $(CSRCS)
	      @ echo Making tags ...
	      @ $(CTAGS) -t $(HHDRS) $(CSRCS)
	      @ sed -e 's/\.c	/.C	/' -e 's/\.h	/.H	/' tags >tags.tmp
	      @ mv tags.tmp tags

depend:		$(HHDRS) $(CSRCS)
	      @ echo Remaking dependencies...
	      @ $(MAKETD) -mmf.dep $(INCL) $(DEFS) $(CSRCS)

FORCE:

kanji/kanji.a:	FORCE
	      @ cd kanji; $(MAKE_KANJI) all

Cstrings/Cstrings.a:	FORCE
	      @ cd Cstrings; $(MAKE_CSTR) all

Css.o:		Css.c
		$(CC) -c $(SHARESTR) $*.c

#if KANJI
.C.c:;	      @ echo processing $*.C ...
	      @ $(KPP) $(FCODE) $*.C
	      @ mv Cxx.c $@
#else
.C.c:;	      @ cp $*.C $@
#endif

.c.o:;		$(CC) $(CFLAGS) -c $*.c

#if KANJI
.H.h:;	      @ echo processing $*.H ...
	      @ $(KPP) $(FCODE) $*.H
	      @ mv Cxx.c $@
#else
.H.h:;	      @ cp $*.H $@
#endif

Css.c:		Cstr.data
	      @ echo Making $@ ...
	      @ $(FTOC)

$(KPP) $(FTOC):	$(LIBS) FORCE
	      @ cd kpp; $(MAKE_KPP) all

co:		$(HDRS) $(SRCS) $(MANUAL) $(DOCS)

.DEFAULT:;	co -l $@

#if KANJI
$(HHDRS) $(CSRCS): $(KPP)
#endif

ebind.h: ebind.H
ecomm.h: ecomm.H
edef.h: edef.H
efunc.h: efunc.H
epath.h: epath.H
estruct.h: estruct.H
evar.h: evar.H
eversion.h: eversion.H

basic.c: basic.C
bind.c: bind.C
buffer.c: buffer.C
crypt.c: crypt.C
display.c: display.C
eval.c: eval.C
exec.c: exec.C
file.c: file.C
fileio.c: fileio.C
input.c: input.C
isearch.c: isearch.C
kanji.c: kanji.C
line.c: line.C
main.c: main.C
random.c: random.C
region.c: region.C
search.c: search.C
spawn.c: spawn.C
tcap.c: tcap.C
termio.c: termio.C
version.c: version.C
window.c: window.C
word.c: word.C
