/*
 *	Spawn:	various OS access commands
 *		for MicroEMACS
 */

#include	"ecomm.h"
#include	"estruct.h"
#include	"edef.h"

#include	<signal.h>
extern int vttidy();

/*
 * Create a subjob with a copy of the command intrepreter in it. When the
 * command interpreter exits, mark the screen as garbage so that you do a full
 * repaint. Bound to "^X C". The message at the start in VMS puts out a newline.
 * Under some (unknown) condition, you don't get one free when DCL starts up.
 */
/*ARGSUSED*/
spawncli(f, n)
{
	register char *cp;
	char	*getenv();

#if	RESTRICT
	/* don't allow this command if restricted */
	if (restflag)
		return(resterr());
#endif

	movecursor(term.t_nrow, 0);		/* Seek to last line.	*/
	TTflush();
	if ((cp = getenv("SHELL")) != NULL && *cp != '\0')
		(void)System(cp);
	else
#if	BSD
		(void)System("exec /bin/csh");
#endif
#if	USG
		(void)System("exec /bin/sh");
#endif
	sgarbf = TRUE;
	sleep(2);
	return(TRUE);
}

#if	BSD

bktoshell()		/* suspend MicroEMACS and wait to wake up */
{
	vttidy();
	(void)kill(0, SIGTSTP); /* suspend all processes within a group */
	TTkopen();
}

rtfrmshell()
{
	ttopen();
	curwp->w_flag = WFHARD;
	sgarbf = TRUE;
}
#endif /*BSD*/

/*
 * Run a one-liner in a subjob. When the command returns, wait for a single
 * character to be typed, then mark the screen as garbage so a full repaint is
 * done. Bound to "C-X !".
 */
/*ARGSUSED*/
spawn(f, n)
{
	register int	s;
	Char		line[NLINE];

#if	RESTRICT
	/* don't allow this command if restricted */
	if (restflag)
		return(resterr());
#endif

	if ((s=mlreply(CSTR("!"),
				  line, NLINE)) != TRUE)
		return (s);
	TTputc('\n');                /* Already have '\r'    */
	TTflush();
	(void)System(cfromC(line));
	mlputs(
#if KMSGS
	    kterminal? CSTR("[$B=*N;(J]"):
#endif
	    CSTR("[End]"));
	TTflush();
	/* Pause. */
	while ((s = tgetc()) != '\r' && s != '\n' && s != ' ')
		;
	sgarbf = TRUE;
	return (TRUE);
}

/*
 * Pipe a one line command into a window
 * Bound to ^X @
 */
/*ARGSUSED*/
pipecmd(f, n)
{
	register int	s;	/* return status from CLI */
	register WINDOW *wp;	/* pointer to new window */
	register BUFFER *bp;	/* pointer to buffer to zot */
	Char	line[NLINE];	/* command line send to shell */
	static Char *bname = CSTR("[command]");
	static Char *filnam = CSTR(",command");

#if	RESTRICT
	/* don't allow this command if restricted */
	if (restflag)
		return(resterr());
#endif

	/* get the command to pipe in */
	if ((s=mlreply(CSTR("@"),
				  line, NLINE)) != TRUE)
		return(s);

	/* get rid of the command output buffer if it exists */
	if ((bp=bfind(bname, FALSE, 0)) != FALSE) {
		/* try to make sure we are off screen */
		wp = wheadp;
		while (wp != NULL) {
			if (wp->w_bufp == bp) {
				(void)onlywind(FALSE, 1);
				break;
			}
			wp = wp->w_wndp;
		}
		if (zotbuf(bp) != TRUE)
			return(FALSE);
	}

	TTflush();
	(void)Cstrcat(line,CSTR(">"));
	(void)Cstrcat(line,filnam);
	(void)System(cfromC(line));
	TTflush();
	sgarbf = TRUE;
	s = TRUE;

	if (s != TRUE)
		return(s);

	/* split the current window to make room for the command output */
	if (splitwind(FALSE, 1) == FALSE)
			return(FALSE);

	/* and read the stuff in */
	if (getfile(filnam, FALSE) == FALSE)
		return(FALSE);

	/* set name and make this window in VIEW mode, update all mode lines */
	(void)copystr(&curwp->w_bufp->b_bname, bname);
	(void)copystr(&curwp->w_bufp->b_fname, CSTR(""));
	curwp->w_bufp->b_mode |= MDVIEW;
	upmode();

	/* and get rid of the temporary file */
	(void)unlink(cfromC(filnam));
	return(TRUE);
}

/*
 * filter a buffer through an external DOS program
 * Bound to ^X #
 */
/*ARGSUSED*/
filter(f, n)
{
	register int	s;	/* return status from CLI */
	register BUFFER *bp;	/* pointer to buffer to zot */
	Char line[NLINE];	/* command line send to shell */
	Char *tmpnam;		/* place to store real file name */
	static Char *filnam1 = CSTR(",fltinp");
	static Char *filnam2 = CSTR(",fltout");

#if	RESTRICT
	/* don't allow this command if restricted */
	if (restflag)
		return(resterr());
#endif

	if (curbp->b_mode & MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/

	/* get the filter name and its args */
	if ((s=mlreply(CSTR("#"),
				  line, NLINE)) != TRUE)
		return(s);

	/* setup the proper file names */
	bp = curbp;
	tmpnam = bp->b_fname;	/* save the original name */
	bp->b_fname = NULL;

	/* write it out, checking for errors */
	if (writeout(filnam1) != TRUE) {
		mlwrite(
#if KMSGS
		    kterminal? CSTR("[$B%U%#%k%?MQ%U%!%$%k$K=q9~$a$^$;$s(J]"):
#endif
		    CSTR("[Cannot write filter file]"));
		free((char *)bp->b_fname);
		bp->b_fname = tmpnam;
		return(FALSE);
	}

	TTflush();
	(void)Cstrcat(line, CSTR(" <"));
	(void)Cstrcat(line, filnam1);
	(void)Cstrcat(line, CSTR(" >"));
	(void)Cstrcat(line, filnam2);
	(void)System(cfromC(line));
	TTflush();
	sgarbf = TRUE;

	/* on failure, escape gracefully */
	if (readin(filnam2, FALSE) != TRUE) {
		mlwrite(
#if KMSGS
		    kterminal? CSTR("[$B<B9T7k2L$rFI$_9~$a$^$;$s(J]"):
#endif
		    CSTR("[Can't read the result]"));
		free((char *)bp->b_fname);
		bp->b_fname = tmpnam;
		(void)unlink(cfromC(filnam1));
		(void)unlink(cfromC(filnam2));
		return(s);
	}

	/* reset file name */
	free((char *)bp->b_fname);
	bp->b_fname =  tmpnam;	/* restore name */
	bp->b_flag |= BFCHG;		/* flag it as changed */

	/* and get rid of the temporary file */
	(void)unlink(cfromC(filnam1));
	(void)unlink(cfromC(filnam2));
	return(TRUE);
}

/*
 * execute shell command.
 */
int
System(cmd)
    char *cmd;
{
    int s;
#if	CLOCK
    SIGRET_T (*f)();
    unsigned t;
#endif

#if	CLOCK
    f = signal(SIGALRM, SIG_DFL);
    t = alarm((unsigned)0);
#endif
    TTclose();
    s = system(cmd);
    ttopen();
#if	CLOCK
    (void)signal(SIGALRM, f);
    (void)alarm(t);
#endif
    return s;
}
