/*
 * The functions in this file negotiate with the operating system for
 * characters, and write characters in a barely buffered fashion on the display.
 * All operating systems.
 */
#include	"ecomm.h"
#include	"estruct.h"
#include	"edef.h"

static Char *dummy = CSTR("");

#if	USG			/* System V */
# include	<signal.h>
# include	<termio.h>
struct	termio	otermio;	/* original terminal characteristics */
struct	termio	ntermio;	/* charactoristics to use inside */
#endif

#if	BSD
# undef	CTRL
# include	<sgtty.h>	 /* for stty/gtty functions */
# include	<signal.h>
struct	sgttyb	ostate;		 /* saved tty state */
struct	sgttyb	nstate;		 /* values for editor mode */
struct tchars	otchars;	/* Saved terminal special character set */
struct tchars	ntchars = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
				/* A lot of nothing */
# include <sys/ioctl.h>		/* to get at the typeahead */
extern	SIGRET_T rtfrmshell();	/* return from suspended shell */
#if BSD_4_2W
extern	SIGRET_T change_win_size();	/* change window size & redisplay */
#endif
# define TBUFSIZ 128
char tobuf[TBUFSIZ];		/* terminal output buffer */
#endif /*BSD*/

/*
 * This function is called once to set up the terminal device streams.
 */
ttopen()
{
#if	USG
	ioctl(0, TCGETA, &otermio);	/* save old settings */
	ntermio.c_iflag = 0;		/* setup new settings */
	ntermio.c_oflag = 0;
	ntermio.c_cflag = otermio.c_cflag;
	ntermio.c_lflag = 0;
	ntermio.c_line = otermio.c_line;
	ntermio.c_cc[VMIN] = 1;
	ntermio.c_cc[VTIME] = 0;
	ioctl(0, TCSETAW, &ntermio);	/* and activate them */
#endif

#if	BSD
	(void)gtty(0, &ostate);			/* save old state */
	(void)gtty(0, &nstate);			/* get base of new state */
	nstate.sg_flags |= RAW;
	nstate.sg_flags &= ~(ECHO|CRMOD);	/* no echo for now... */
	(void)stty(0, &nstate);			/* set mode */
	(void)ioctl(0, (int)TIOCGETC, (char *)&otchars); /* Save old characters */
	(void)ioctl(0, (int)TIOCSETC, (char *)&ntchars); /* Place new character into K */
	/* provide a smaller terminal output buffer so that
	   the type ahead detection works better (more often) */
	setbuffer(stdout, &tobuf[0], TBUFSIZ);
	(void)signal(SIGTSTP,SIG_DFL);		/* set signals so that we can */
	(void)signal(SIGCONT,rtfrmshell);	/* suspend & restart emacs */
#if BSD_4_2W
	(void)signal(SIGWINCH,change_win_size);	/* change window size */
#endif
#endif	/*BSD*/
}

/*
 * This function gets called just before we go back home to the command
 * interpreter.
 */
ttclose()
{
#if	USG
	ioctl(0, TCSETAW, &otermio);	/* restore terminal settings */
#endif

#if	BSD
	(void)stty(0, &ostate);
	(void)ioctl(0, (int)TIOCSETC, (char *)&otchars); /* Place old character into K */
#endif
}

/*
 * Write a character to the display. On VMS, terminal output is buffered, and
 * we just put the characters in the big array, after checking for overflow.
 */
ttputc(c)
    Char c;
{
#if	KANJI
	kkputc(c, stdout);
#else
	putchar(c);
#endif /*KANJI*/
}

/*
 * Flush terminal buffer. Does real work where the terminal output is buffered
 * up. A no-operation on systems where byte at a time terminal I/O is done.
 */
ttflush()
{
#if KANJI
  kkputc(EOF, stdout);
#endif
  (void)fflush(stdout);
}

/*
 * Read a character from the terminal, performing no editing and doing no echo
 * at all.
 */
Char
ttgetc()
{
#if KANJI
	Char kkgetc();

	return(kkgetc(stdin));
#else
	return (Char)getchar();
#endif
}

#if	TYPEAH
/* typahead:	Check to see if any characters are already in the
		keyboard buffer
*/
typahead()
{
#if	BSD
	int x;	/* holds # of pending chars */

	return((ioctl(0, (int)FIONREAD, (char *)&x) < 0) ? 0 : x);
#endif
#if	USG
	return 0;
#endif	/*BSD*/
}
#endif	/*TYPEAH*/
